//****************************************************************************
// Copyright (c) 2001 Microsoft Corporation.
//
// @File: ExpressionTask.cs
//
// Purpose: 
//  
// Notes:
//
// History:
//
//     @Version:
//     140123 KXH  04/25/03 Implement validation in Expression task.
//     139040 KXH  04/23/03 A task for evaluating expressions.  This is a 
//                          utility task and is
//
// @EndHeader@
//****************************************************************************
using System;
using System.Xml;
using System.Threading;
using Microsoft.SqlServer.Dts.Runtime;


namespace SSIS2K5.Samples.Tasks.ExpressionTask
{
	#region Attributes
	// These attributes allow the runtime to enumerate the task and associate the task's UI
	// with the task.  The UITypeName must match the name of the task UI in the Expression namespace.
	[DtsTask(DisplayName = "Expression Task",
		Description = "Evaluates DTS Expressions",
		TaskType = "DTS90",
    UITypeName = "SSIS2K5.Samples.Tasks.ExpressionTask.ExpressionTaskUI, SSIS2K5.Samples.ExpressionTaskUI, Version=1.0.0.1, Culture=Neutral, PublicKeyToken=125b07dfcbe5fa19",
		IconResource = "SSIS2K5.Samples.Tasks.ExpressionTask.TaskIcon.ico",
     TaskContact = "Expression Task; Microsoft SQL Server Integration Services 2005, By Kirk Haselden; For Version 9.0 of Integration Services; (c) 2006 Kirk Haselden;http://www.CustomTask.com/ExpressionTask"),
		System.Runtime.InteropServices.Guid("730F5F89-3336-4568-B2D1-7E8A26E27AAA")]
	#endregion

	#region Task
	public class ExpressionTask : Task, IDTSBreakpointSite, IDTSComponentPersist
	{		
		#region Internals

		// The expression evaluator
		
		// The breakpoint manager we get when first loaded from AcceptBreakpointManager()
		private BreakpointManager m_bpManager = null;
		private ExpressionEval m_ExpressionEval = null;

		// Suspend flags for the interface.  These are int instead of bool so that they can
		// be changed by Interlock and are therefore thread safe.
		private int m_debugMode = 0;
		private int m_suspendRequired = 0;

		// Events to indicate when we can execute and when we are suspended.  m_canExecute is
		// only modified by call so the IDTSBreakpointSite interface.  m_suspended is only
		// modified by calls to Execute().
		private ManualResetEvent m_suspended = new ManualResetEvent(true);
		private ManualResetEvent m_canExecute = new ManualResetEvent(true);
		private static ManualResetEvent m_WriteCompletionEvent = new ManualResetEvent(false);

		// Breakpoint IDs
		private const int BPID1 = 1;	

		// Execution Result
		private object m_ExecVal = null;

		// Simplified method for logging
		private void WriteLog(IDTSLogging log, string message)
		{
			if (log != null)
			{
				byte [] nothing = null;
				log.Log("Status", null, null, null, null, null, message, DateTime.Now, DateTime.Now, 0, ref nothing);
			}
		}

		static void ChangeEvent(ManualResetEvent e, bool shouldSet)
		{
			bool succeeded;

			if (shouldSet)
				succeeded = e.Set();
			else
				succeeded = e.Reset();

			if (!succeeded)
				throw new Exception("Synchronization object failed.");
		}

		public ExpressionTask()
		{
			m_ExpressionEval = new ExpressionEval();	
		}


		#endregion

		#region ExpressionProperties
		
		// Path to watch for new files
		private string m_Expression;
		public string Expression
		{
			get {return (m_Expression == null ? "" : m_Expression);}
			set 
			{
				m_Expression = value;
				if( m_ExpressionEval != null )
					m_ExpressionEval.Expression = value;
			}
		}

		public string ValidateExpression( VariableDispenser variableDispenser )
		{
			string s = "Expression is valid";
			try
			{			
				if( m_ExpressionEval != null )
					m_ExpressionEval.Validate(DtsConvert.ToVariableDispenser90(variableDispenser)); 
			}
			catch( Exception e )
			{
				s = e.Message;
			}
			return s;
		}

		public string Evaluate(VariableDispenser variableDispenser)
		{
			string retValue = string.Empty;

			try
			{
				if (m_ExpressionEval != null)
				{
					object evaluatedValue = m_ExpressionEval.Evaluate(DtsConvert.ToVariableDispenser90(variableDispenser));

					if (evaluatedValue != null)
					{
						retValue = evaluatedValue.ToString();
					}
				}
			}
			catch (Exception ex)
			{
				retValue = ex.Message;
			}

			return retValue;
		}
		#endregion

		#region IDTSExecutable
		  public override Microsoft.SqlServer.Dts.Runtime.DTSExecResult Execute(Microsoft.SqlServer.Dts.Runtime.Connections connections, 
			  Microsoft.SqlServer.Dts.Runtime.VariableDispenser variableDispenser, 
			  IDTSComponentEvents events,
			  IDTSLogging log, 
			  object transaction)
		  {
			  Microsoft.SqlServer.Dts.Runtime.DTSExecResult execResult = DTSExecResult.Success;
			  try
			  {
				  // Create an Expression Evaluator
				  m_ExpressionEval.Expression = m_Expression;

				  // While a task is not executing it is suspended.  Now that we are executing,
				  // change to not suspended.
				  ChangeEvent(m_suspended, false);

				  // Check for a suspend before doing any work in case the suspend and execute calls
				  // were initiated at virtually the same time.
				  CheckAndSuspend();
    
				  // If the package wants us to stop executing, quit
				  if (true == events.FireQueryCancel())
					  return DTSExecResult.Canceled;

				  // This is a point in the code where we hit a breakpoint
				  CheckAndFireBreakpoint(events, BPID1);

				  // Evaluate the expression
				  m_ExpressionEval.Expression = m_Expression;				  
				  m_ExpressionEval.Evaluate( DtsConvert.ToVariableDispenser90(variableDispenser));


				  // Write to the log
				  WriteLog(log, "The file name was: " + m_ExecVal);
		            
				  // This is a good suspend point but not one we expose a breakpoint on
				  CheckAndSuspend();

				  // Do even more task stuff.
				  // If the package wants us to stop executing, quit
				  if (true == events.FireQueryCancel())
					  return DTSExecResult.Canceled;

				  // Go to a suspended mode when exiting so that a suspend request that arrives
				  // right now does not get blocked unnecessarily
				  ChangeEvent(m_suspended, true);
				  			  }
			  catch (Exception e)
			  {
				  execResult = DTSExecResult.Failure;
				  if (events != null)
					  events.FireError(0, e.Source, e.Message, "Expression Task", 0);
			  }
			  return execResult;
		  }

		  public override Microsoft.SqlServer.Dts.Runtime.DTSExecResult Validate(Microsoft.SqlServer.Dts.Runtime.Connections connections, 
			  Microsoft.SqlServer.Dts.Runtime.VariableDispenser variableDispenser, 
			  IDTSComponentEvents events,
			  IDTSLogging log)
		  {
			  if (m_Expression.Length == 0)
				  return DTSExecResult.Success;
		
			  DTSExecResult execResult = DTSExecResult.Success;
			  try
			  {
				  m_ExpressionEval.Expression = m_Expression;
				  m_ExpressionEval.Validate(DtsConvert.ToVariableDispenser90(variableDispenser));
			  }
			  catch (Exception e)
			  {
				  execResult = DTSExecResult.Failure;
				  if (events != null)
					  events.FireError(0, e.Source, e.Message, "Expression Task", 0);
			  }
		
			  return execResult;
		  }
	#endregion

		#region IDTSTask overrides
		  /// <summary>
		  /// The execution value is assigned the name of the new file
		  /// </summary>
		  public override object ExecutionValue
		  {
			  get {return m_ExecVal;}
		   }
	#endregion

		#region IDTSBreakpointSite
		   public void SuspendExecution()
		   {
			   // This lock is also taken by the set SuspendRequired method.  It prevents this
			   // call from override an aborted suspension.  See comments in set SuspendRequired.
			   lock (this)
			   {
				   // If a suspend is required, do it.
				   if (m_suspendRequired != 0)
					   ChangeEvent(m_canExecute, false);
			   }

			   // We can't return from Suspend until the task is "suspended".  This can happen
			   // one of two ways: the m_suspended event occurs, indicating that the execute thread
			   // has suspended or the canExecute flag is set, indicating that a suspend is no
			   // longer required.
			   WaitHandle [] suspendOperationComplete = {m_suspended, m_canExecute};
			   WaitHandle.WaitAny(suspendOperationComplete);
		   }

		   public void ResumeExecution()
		   {
			   ChangeEvent(m_canExecute, true);
		   }

		   public bool SuspendRequired
		   {
			   get {return m_suspendRequired != 0;}
				set
				{
					// This lock is also taken by Suspend().  Since it is possible for the package to be
					// suspended and resumed in quick succession, this property put might happen before
					// the actual Suspend() call.  Without the lock, the Suspend() might reset the canExecute
					// event after we set it to abort the suspension.
					lock (this)
					{
						Interlocked.Exchange(ref m_suspendRequired, value ? 1 : 0);

						if (!value)
							ResumeExecution();
					}
				}
			}

			public void AcceptBreakpointManager(BreakpointManager manager)
			{
				m_bpManager = manager;
				manager.CreateBreakpointTarget(BPID1, "Break right before evaluating the expression");
				
			}

			public bool DebugMode
			{
				get {return m_debugMode != 0;}
				 set {Interlocked.Exchange(ref m_debugMode, value ? 1 : 0);}
			  }

			  private void CheckAndSuspend()
			  {
				  // Loop until we can execute.  The loop is required rather than a simple if
				  // because there is a time between the return from WaitOne and the reset that we
				  // might receive another Suspend call.  Suspend() will see that we are suspended
				  // and return.  So we need to rewait.
				  while (!m_canExecute.WaitOne(0, false))
				  {
					  ChangeEvent(m_suspended, true);
					  m_canExecute.WaitOne();
					  ChangeEvent(m_suspended, false);
				  }
			  }

			  private void CheckAndFireBreakpoint(IDTSComponentEvents events, int breakpointID)
			  {
				  // If the breakpoint is enabled, fire it.
				  if (m_debugMode != 0 && m_bpManager.IsBreakpointTargetEnabled(breakpointID))
				  {
					  // Enter a suspend mode before firing the breakpoint.  Firing the breakpoint
					  // will cause the runtime to call Suspend on this task.  Since we are blocked
					  // on the breakpoint, we are suspended, so indicate that.
					  ChangeEvent(m_suspended, true);
					  events.FireBreakpointHit(m_bpManager.GetBreakpointTarget(breakpointID));
					  ChangeEvent(m_suspended, false);
				  }

				  // Check for a suspension for two reasons: 1. If we are at a point where we could
				  // fire a breakpoint, we are at a valid suspend point.  Even if we didn't hit a
				  // breakpoint, the runtime may have called suspend, so check for it.  2. Between
				  // the return from OnBreakpointHit and the reset of the event, it is possible to have
				  // received a suspend call that we returned from because we were suspended.  We need
				  // to be sure it is ok to continue executing now.
				  CheckAndSuspend();
			  }
		#endregion	

		#region IDTSComponentPersist
			  public void SaveToXML(XmlDocument doc, IDTSInfoEvents events)
			  {
				  XmlNode node = doc.CreateNode(XmlNodeType.Element, "ExpressionTask", "");
				  XmlElement element = node as XmlElement;
				  doc.AppendChild(node);

				  XmlAttribute attr = doc.CreateAttribute("Expression");
				  attr.Value = m_Expression;
				  element.Attributes.Append(attr);			   
			  }

			  public void LoadFromXML(XmlElement root, IDTSInfoEvents events)
			  {
				  XmlNode attr = root.Attributes.GetNamedItem("Expression");
				  m_Expression = attr.Value;
			  }
		#endregion
      			
		  }
	#endregion
}