using System;
using Microsoft.SqlServer.Dts.Runtime;
using Microsoft.SqlServer.Dts.Pipeline;
using Microsoft.SqlServer.Dts.Runtime.Wrapper;
using Microsoft.SqlServer.Dts.Pipeline.Wrapper;
using System.Data;
using System.Diagnostics;
using System.Data.Odbc;
using System.Data.Common;
using System.Text;

using System.Windows.Forms;

namespace SSIS2K8.Samples.Transforms.ODBCDest
{
	[ DtsPipelineComponent(
		DisplayName = "ADONET ODBC Destination",
		Description = "An ADO.NET ODBC Destination",
		IconResource = "SSIS2K8.Samples.Transforms.ODBCDest.Component.ico",
		ComponentType = ComponentType.DestinationAdapter ) ]
	public class ODBCDest : PipelineComponent
   {

      #region PRIVATE CONSTANTS
      private const string CONNECTION_NAME = "IDbConnection";
      private const string OUTPUT_NAME = "ADONET Output";
      private const string TABLE_NAME = "Output Table Name";
      private const string CATALOG_NAME = "Output Catalog Name";
      private const int MAX_NUMERIC_PRECISION = 38;
      private const int DTS_PIPELINE_CTR_ROWSWRITTEN = 103;
      private const int DTS_PIPELINE_CTR_BLOBBYTESWRITTEN = 118;
      const string INSERT = "INSERT INTO ";
      const string VALUES = " VALUES (";
      #endregion

      #region PRIVATE MEMBERS
      private bool m_connected = false;
      private int m_countMetaDataColumns = 0;
      private int m_errorOutID;
      private System.Data.Odbc.OdbcConnection m_conn = null;
      private DTSRowDisposition m_rowdisp = DTSRowDisposition.RD_FailComponent;
      private OdbcDataAdapter m_odbcAdapter = null;
      private int[] m_cols;
      DataTable m_batchTable;
      DataTable m_metaDataTable;
      private int[] m_idxColumn;
      #endregion

      #region PRIVATE METHODS
      /// Set the component version to match the assembly version.      
      private void SetComponentVersion()
      {         
         // Get the assembly version and set that as our current version.
         DtsPipelineComponentAttribute attr = (DtsPipelineComponentAttribute)
               Attribute.GetCustomAttribute(this.GetType(), typeof(DtsPipelineComponentAttribute), false);
         ComponentMetaData.Version = attr.CurrentVersion;
      }

      // Get the Connection Manager object
      private IDTSConnectionManager100 GetConnectionManager()
      {
         IDTSRuntimeConnection100 conn = ComponentMetaData.RuntimeConnectionCollection[CONNECTION_NAME];
         if (conn == null)
         {
            bool bCancel;
            ComponentMetaData.FireError(2, "",
               "Cannot find the runtime Connection Manager" +
               conn.ConnectionManagerID, "", 0, out bCancel);
            throw new Exception("No runtime connection");
         }

         // Get the connection manager from the runtime connection
         IDTSConnectionManager100 connmgr = conn.ConnectionManager;

         if (connmgr == null)
         {
            bool bCancel;
            ComponentMetaData.FireError(2, "",
               "Cannot find the runtime Connection Manager" +
               conn.ConnectionManagerID, "", 0, out bCancel);
            throw new Exception("No runtime connection");
         }

         return connmgr;
      }

      // Get the physical connection
      private OdbcConnection GetConnection(object tx)
      {
         // Get the connection manager from the runtime connection
         IDTSConnectionManager100 connmgr = GetConnectionManager();

         // Get the connection from the Connection Manager
         OdbcConnection conn = (connmgr.AcquireConnection(tx)) as OdbcConnection;
         if (conn == null)
         {
            bool bCancel;
            ComponentMetaData.FireError(3, "", "Unable to retrieve connection", "", 0, out bCancel);
         }
         return conn;
      }

      // Retrieve the table name custom property
      private string GetTableNameCustomProperty()
      {
         // Get the table name
         IDTSCustomProperty100 propTableName = ComponentMetaData.CustomPropertyCollection[TABLE_NAME];
         string strTN = (propTableName.Value == null ? null : propTableName.Value.ToString());

         // Is it empty
         if (strTN == null || strTN.Trim().Length == 0)
         {
            // Must have a output table name. 
            bool bCancel;
            ComponentMetaData.FireError(4, "", "No destination table name provided", "", 0, out bCancel);
            throw new Exception("Destination table not provided");
         }
         return strTN;
      }

      // Retrieve the table name custom property
      private string GetCatalogNameCustomProperty()
      {
         // Get the table name
         IDTSCustomProperty100 propCatalogName = ComponentMetaData.CustomPropertyCollection[CATALOG_NAME];
         string strCN = (propCatalogName.Value == null ? null : propCatalogName.Value.ToString());

         // Is it empty
         if (strCN == null || strCN.Trim().Length == 0)
         {
            // Must have a output table name. 
            bool bCancel;
            ComponentMetaData.FireError(4, "", "No destination catalog name provided", "", 0, out bCancel);
            throw new Exception("Destination catalog not provided");
         }
         return strCN;
      }


      // Get a table to hold the rows temporarily until it contains BatchSize rows
      // Then we'll update the destination in bigger batches.
      private void CreateBatchTable(string strTableName)
      {
         try
         {
            // Get the Schema into m_metaDataTable;
            CreateSchemaTable(strTableName);

            // Create new data table
            m_batchTable = new DataTable();

            // Build the table MetaData. 
            int countColumns = m_metaDataTable.Rows.Count;

            // Preload with columns
            DataColumn[] columns = new DataColumn[countColumns];

            for (int iColumn = 0; iColumn < countColumns; iColumn++)
            {
               // Point to the row describing this column
               DataRow rowCurr = m_metaDataTable.Rows[iColumn];

               // Build the columns array in the order found in the table.
               // Makes it easier on the user and the programmer.
               columns[Convert.ToInt16(rowCurr["ORDINAL_POSITION"]) - 1] = new DataColumn(
                  rowCurr["COLUMN_NAME"].ToString(),
                  GetOdbcDbTypeFromName(rowCurr["DATA_TYPE"].ToString()).GetType());

            }
            m_batchTable.Columns.AddRange(columns);

         }
         catch (Exception ex)
         {
            bool bCancel;
            ComponentMetaData.FireError(51, "ODBCDest", "Failed to get batch table",
               "", 0, out bCancel);
            throw ex;
         }
      }

      private void CreateSchemaTable(string strTableName)
      {
         try
         {
            // Create the table to hold the table metadata
            m_metaDataTable = new DataTable();

            // Get the InitialCatalog. The Catalog is not stored on the Connection Manager
            // like for other ADO.Net connections, so have to explicitly add it in Advanced UI.
            string initialCatalog = GetCatalogNameCustomProperty();
            IDTSConnectionManager100 cm = GetConnectionManager();

            // Build the restrictions. We just want to get column information for the table.
            // In the Initial Catalog.
            string[] restrictions = new string[4] { initialCatalog, null, strTableName, null };

            // Get the schema.
            m_metaDataTable = m_conn.GetSchema("Columns", restrictions);
         }
         catch (Exception ex)
         {
            bool bCancel;
            ComponentMetaData.FireError(50, "ODBCDest",
               "Failed to retrieve table metadata", "", 0, out bCancel);
            throw ex;
         }
      }

      private static bool TranslateOdbcTypeToDts(string typeName, ref DataType dtsColType)
      {
         switch (typeName)
         {
            case "binary": dtsColType = DataType.DT_BYTES; break;
            case "bit": dtsColType = DataType.DT_BOOL; break;
            case "char": dtsColType = DataType.DT_STR; break;
            case "datetime": dtsColType = DataType.DT_DBTIMESTAMP; break;
            case "decimal": dtsColType = DataType.DT_NUMERIC; break;
            case "float": dtsColType = DataType.DT_R8; break;
            case "image": dtsColType = DataType.DT_IMAGE; break;
            case "bigint": dtsColType = DataType.DT_I8; break;
            case "int": dtsColType = DataType.DT_I4; break;
            case "money": dtsColType = DataType.DT_CY; break;
            case "nchar": dtsColType = DataType.DT_WSTR; break;
            case "ntext": dtsColType = DataType.DT_NTEXT; break;
            case "numeric": dtsColType = DataType.DT_NUMERIC; break;
            case "nvarchar": dtsColType = DataType.DT_WSTR; break;
            case "real": dtsColType = DataType.DT_R4; break;
            case "smalldatetime": dtsColType = DataType.DT_DBTIMESTAMP; break;
            case "smallint": dtsColType = DataType.DT_I2; break;
            case "smallmoney": dtsColType = DataType.DT_CY; break;
            case "sysname": dtsColType = DataType.DT_STR; break;
            case "text": dtsColType = DataType.DT_TEXT; break;
            case "timestamp": dtsColType = DataType.DT_BYTES; break;
            case "tinyint": dtsColType = DataType.DT_UI1; break;
            case "uniqueidentifier": dtsColType = DataType.DT_GUID; break;
            case "varbinary": dtsColType = DataType.DT_BYTES; break;
            case "varchar": dtsColType = DataType.DT_STR; break;
            case "xml": dtsColType = DataType.DT_WSTR; break;

            default:
               return false;
         }
         return true;
      }



      private static OdbcType GetOdbcDbTypeFromName(string typeName)
      {
         switch (typeName)
         {
            case "binary": return OdbcType.Binary;
            case "bit": return OdbcType.Bit;
            case "char": return OdbcType.Char;
            case "datetime": return OdbcType.DateTime;
            case "decimal": return OdbcType.Decimal;
            case "float": return OdbcType.Double;
            case "image": return OdbcType.Image;
            case "bigint": return OdbcType.BigInt;
            case "int": return OdbcType.Int;
            case "money": return OdbcType.Double;
            case "nchar": return OdbcType.NChar;
            case "ntext": return OdbcType.NText;
            case "numeric": return OdbcType.Double;
            case "nvarchar": return OdbcType.NVarChar;
            case "real": return OdbcType.Real;
            case "smalldatetime": return OdbcType.SmallDateTime;
            case "smallint": return OdbcType.SmallInt;
            case "smallmoney": return OdbcType.Real;
            case "sysname": return OdbcType.NVarChar;
            case "text": return OdbcType.Text;
            case "timestamp": return OdbcType.Timestamp;
            case "tinyint": return OdbcType.TinyInt;
            case "uniqueidentifier": return OdbcType.UniqueIdentifier;
            case "varbinary": return OdbcType.VarBinary;
            case "varchar": return OdbcType.VarChar;
            case "xml": return OdbcType.Text;
            default: return OdbcType.Text;
         }
      }
      #endregion

      #region DESIGNTIME METHODS
      // Get and store the connection
      public override void AcquireConnections(object transaction)
      {
         m_conn = GetConnection(transaction);

         // We're connected.
         if( null != m_conn )
            m_connected = true;
      }

      // Release connection
      public override void ReleaseConnections()
      {
         if (m_conn != null)
         {
            IDTSRuntimeConnection100 conn;

            // Get the runtime connection
            conn = ComponentMetaData.RuntimeConnectionCollection[CONNECTION_NAME];

            // Close it, or the subsequent release by the provider will throw.
            if (m_conn is OdbcConnection)
            {
               m_conn.Close();
            }
            else
            {
               IDTSConnectionManager100 connmgr = conn.ConnectionManager;
               connmgr.ReleaseConnection(m_conn);
            }
         }
         m_conn = null;
         m_connected = false;                  
      }

      public override void ProvideComponentProperties()
      {
         // Start out clean, remove anything put on by the base class
         RemoveAllInputsOutputsAndCustomProperties();

         //
         // Setup component information
         //
         // Get the assembly version and set that as our current version.
         SetComponentVersion();

         // Name the component and add component information
         ComponentMetaData.Name = "ODBCDest";
         ComponentMetaData.Description = "Sample ADO.Net ODBC Destination Adapter";
         ComponentMetaData.ContactInfo = ComponentMetaData.Description +
            ";Microsoft SQL Server Integration Services 2005, By Kirk Haselden;" +
            " (c) 2006 Kirk Haselden" +
            ComponentMetaData.Version.ToString();         

         // Get the runtime connection collection.
         IDTSRuntimeConnectionCollection100 pIDTSRuntimeConnectionCollection = ComponentMetaData.RuntimeConnectionCollection;

         // See if there	is already a runtime object.
         IDTSRuntimeConnection100 pIDTSRuntimeConnection;
         try
         {
            pIDTSRuntimeConnection = pIDTSRuntimeConnectionCollection[CONNECTION_NAME];
         }
         catch (Exception)
         {
            // Must not be there, make one
            pIDTSRuntimeConnection = pIDTSRuntimeConnectionCollection.New();
            pIDTSRuntimeConnection.Name = CONNECTION_NAME;
            pIDTSRuntimeConnection.Description = "Connection for the ADONET ODBC Destination";
         }

         // 
         // Setup input and external metadata 
         //

         // Add an input
         IDTSInput100 inp = ComponentMetaData.InputCollection.New();

         // Name the input and mention that it has side effects
         inp.Name = "Input";
         inp.HasSideEffects = true;

         // Put error disposition info on the input
         inp.ErrorOrTruncationOperation = "Error or Truncation";
         inp.ErrorRowDisposition = DTSRowDisposition.RD_FailComponent;

         // We use external metadata on that input.
         inp.ExternalMetadataColumnCollection.IsUsed = true;

         // Add the error output
         IDTSOutput100 outError = ComponentMetaData.OutputCollection.New();
         outError.Name = "Error Output";
         outError.IsErrorOut = true;
         outError.SynchronousInputID = ComponentMetaData.InputCollection[0].ID;
         outError.ExclusionGroup = 1;

         // Add the TableName property and set it as expressionable
         IDTSCustomProperty100 propTableName = ComponentMetaData.CustomPropertyCollection.New();
         propTableName.Name = TABLE_NAME;
         propTableName.Description = "The name of the table to insert into";
         propTableName.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;

         // Add the CatalogName property and set it as expressionable
         IDTSCustomProperty100 propCatalogName = ComponentMetaData.CustomPropertyCollection.New();
         propCatalogName.Name = CATALOG_NAME;
         propCatalogName.Description = "The name of the destination catalog";
         propCatalogName.ExpressionType = DTSCustomPropertyExpressionType.CPET_NOTIFY;

         // This adapter uses error dispositions
         ComponentMetaData.UsesDispositions = true;

         // Set we want to validate external metadata
         ComponentMetaData.ValidateExternalMetadata = true;
      }

      public override void ReinitializeMetaData()
      {
         //Baseclass may have some work to do here
         base.ReinitializeMetaData();

         // Get the input
         IDTSInput100 inputMain = ComponentMetaData.InputCollection[0];

         // If we are in a connected state, then continue.
         if (!m_connected)
            throw new Exception("There is no connection available to create metadata");

         // Get the column info from the metadata table
         // Start out fresh
         inputMain.InputColumnCollection.RemoveAll();
         inputMain.ExternalMetadataColumnCollection.RemoveAll();

         // Get the table metadata into m_metaDataTable.
         CreateSchemaTable(GetTableNameCustomProperty());
       
         // Every row of the table describes one column
         // Now setup the ExternalMetadata Columns
         m_countMetaDataColumns = m_metaDataTable.Rows.Count;
         for (int iColumn = 0; iColumn < m_countMetaDataColumns; iColumn++)
         {
            // Create a new column
            IDTSExternalMetadataColumn100 inputcolNewMain = inputMain.ExternalMetadataColumnCollection.NewAt(iColumn);

            // Point to the row describing this column
            DataRow rowCurr = m_metaDataTable.Rows[iColumn];
            
            // Get the name
            inputcolNewMain.Name = rowCurr["COLUMN_NAME"].ToString();

            int CodePage = 0;

            // Convert the type to a pipeline type          
            DataType dtstype = DataType.DT_EMPTY;
            TranslateOdbcTypeToDts(rowCurr["TYPE_NAME"].ToString(), ref dtstype);

            // Set the length for things where the type does not determine the length
            int Length = 0;
            if (dtstype == DataType.DT_WSTR || dtstype == DataType.DT_STR ||
               dtstype == DataType.DT_BYTES || dtstype == DataType.DT_IMAGE ||
               dtstype == DataType.DT_NTEXT || dtstype == DataType.DT_TEXT)
            {
               Length = (int)rowCurr["COLUMN_SIZE"];
            }

            // For str, we need the code page.
            if (dtstype == DataType.DT_STR)
            {
               // Use the default
               CodePage = -1;
            }

            // Numeric or decimal types get precision and scale
            int Precision = 0;
            int Scale = 0;
            if (dtstype == DataType.DT_NUMERIC || dtstype == DataType.DT_DECIMAL)
            {
               if (rowCurr.IsNull("NUM_PREC_RADIX"))
                  Precision = 29;
               else
                  Precision = (int)Convert.ToInt16(rowCurr["NUM_PREC_RADIX"]);
               if (rowCurr.IsNull("DECIMAL_DIGITS"))
                  Scale = 0;
               else
                  Scale = (int)Convert.ToInt16(rowCurr["DECIMAL_DIGITS"]);

               // a decimal may not have a scale, so we make one up
               if (Scale == 255)
                  Scale = 0;

            }

            // Create input columns with the ideal type for conversion
            inputcolNewMain.CodePage = CodePage;
            inputcolNewMain.DataType = dtstype;
            inputcolNewMain.Length = Length;
            inputcolNewMain.Precision = Precision;
            inputcolNewMain.Scale = Scale;
         }         
      }

      // Don't allow user to add inputs
      public override IDTSInput100 InsertInput(DTSInsertPlacement insertPlacement, int inputID)
      {
         bool bCancel;
         ComponentMetaData.FireError(20, "", "Cannot add inputs to ADONET Destination.", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }


      // Don't allow user to add outputs
      public override IDTSOutput100 InsertOutput(DTSInsertPlacement insertPlacement, int outputID)
      {
         bool bCancel;
         ComponentMetaData.FireError(21, "", "Cannot add outputs to ODBC Destination", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }

      // Don't allow user to delete input
      public override void DeleteInput(int inputID)
      {
         bool bCancel;
         ComponentMetaData.FireError(22, "", "Cannot delete inputs to ODBC Destination", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }

      // Don't allow user to delete outputs
      public override void DeleteOutput(int outputID)
      {
         bool bCancel;
         ComponentMetaData.FireError(23, "", "Cannot delete outputs to ODBC Destination", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }

      // Don't allow insert at
      public override IDTSOutputColumn100 InsertOutputColumnAt(int outputID, int outputColumnIndex, string name, string description)
      {
         bool bCancel;
         ComponentMetaData.FireError(24, "", "Cannot insert outputs to ODBC Destination", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }

      // Don't allow deleting output
      public override void DeleteOutputColumn(int outputID, int outputColumnID)
      {
         bool bCancel;
         ComponentMetaData.FireError(25, "", "Cannot delete outputs to ODBC Destination", "", 0, out bCancel);
         throw new Exception("Invalid operation");
      }
            
      // Validate that the settings on the adapter are correct.
		public override DTSValidationStatus Validate()
		{
         try
         {
            DTSValidationStatus status = base.Validate();
            if (status == DTSValidationStatus.VS_ISCORRUPT)
            {
               return status;
            }

            // Should have 1 input
            if (ComponentMetaData.InputCollection.Count != 1)
            {
               bool bCancel;
               ComponentMetaData.FireError(7, "", "Incorrect number of inputs", "", 0,  out bCancel);
               return DTSValidationStatus.VS_ISCORRUPT;
            }

            // Should have 1 output
            if (ComponentMetaData.OutputCollection.Count != 1)
            {
               bool bCancel;
               ComponentMetaData.FireError(8, "", "Incorrect number of outputs", "", 0, out bCancel);
               return DTSValidationStatus.VS_ISCORRUPT;
            }

            // Disposition must be set
            if (ComponentMetaData.InputCollection[0].ErrorRowDisposition == DTSRowDisposition.RD_NotUsed)
            {
               bool bCancel;
               ComponentMetaData.FireError(9, "", "Incorrect disposition", "", 0, out bCancel);
               return DTSValidationStatus.VS_ISCORRUPT;
            }

            // Get the table name
            IDTSCustomProperty100 propTableName = ComponentMetaData.CustomPropertyCollection[TABLE_NAME];
            string strTableName = propTableName.Value == null ? null : propTableName.Value.ToString();

            // Is it empty
            if (strTableName == null || strTableName.Trim().Length == 0)
            {
               bool bCancel;
               ComponentMetaData.FireError(10, "", "No table specified.", "", 0, out bCancel);
               return DTSValidationStatus.VS_ISCORRUPT;
            }

            if (status == DTSValidationStatus.VS_ISVALID && m_connected &&
                ComponentMetaData.ValidateExternalMetadata)
            {
               // If there are no external metadata columns, then make some
               if (ComponentMetaData.InputCollection[0].ExternalMetadataColumnCollection.Count == 0)
                  return DTSValidationStatus.VS_NEEDSNEWMETADATA;
            }
            return status;
         }
         catch (Exception)
         {
            return DTSValidationStatus.VS_ISCORRUPT;
         }
      }
      #endregion

      #region RUNTIME METHODS
      public override void PreExecute()
      {
         // Get the error output
         m_errorOutID = ComponentMetaData.OutputCollection[0].ID;

         // What are we doing with errors ?
         m_rowdisp = ComponentMetaData.InputCollection[0].ErrorRowDisposition;

         // Get the table name from the custom prop
         string strTableName = GetTableNameCustomProperty();
         
         // Create a new command
         OdbcCommand cmd = m_conn.CreateCommand();
         cmd.CommandType = CommandType.Text;
                  
         // Build the INSERT statement
         // Build the command string
         StringBuilder sb = new StringBuilder(INSERT + strTableName + VALUES);

         // Create the batch table 
         CreateBatchTable(strTableName);
                  
         // Get the number of columns
         m_countMetaDataColumns = m_batchTable.Columns.Count;

         // Create the column indexes
         m_cols = new int[m_countMetaDataColumns];

         // External metadata and batch table column count should be the same
         Debug.Assert( m_batchTable.Columns.Count ==
            ComponentMetaData.InputCollection[0].ExternalMetadataColumnCollection.Count );
                  
         // Do some per-column query work         
         for (int i = 0; i < m_countMetaDataColumns; i++)
         {
            sb.Append("?" + ", ");
                        
            // Point to the row describing this column
            DataRow rowCurr = m_metaDataTable.Rows[i];

            // Get the length for character types
            string strLen = rowCurr["COLUMN_SIZE"].ToString();
            int len = strLen.Length == 0 ? 0 : Convert.ToInt32( strLen );
            
                        
            // Set the parameter with name, type and length.
            cmd.Parameters.Add("@p" + i.ToString(),                  
                                            GetOdbcDbTypeFromName(rowCurr["TYPE_NAME"].ToString()),
                                             len);           

         }
         // Add the closing parenthesis
         // Remove the last comma
         sb.Remove(sb.Length - 2, 2);
         sb.Append(" )");

         // Set the command text to contain the insert statement
         cmd.CommandText = sb.ToString();

         // Ignore returned parameters or rows
         cmd.UpdatedRowSource = UpdateRowSource.None;

         // Create and adapter
         m_odbcAdapter = new OdbcDataAdapter(null, m_conn);

         // Tell the adapter about the command object                  
         m_odbcAdapter.InsertCommand  = cmd;
         m_odbcAdapter.InsertCommand.UpdatedRowSource = UpdateRowSource.None;

         //
         // Bind the batchDataTable to the input buffer columns
         // by matching the LineageID from the Input buffer to the batchTable columns.
         //

         // Create an array of objects large enough to hold one row of data
         int cBTCol = m_batchTable.Columns.Count;

         // make an array to hold the indexes in the buffer of the resultset bound pipeline columns
         m_idxColumn = new int[cBTCol];

         // Populate this array with indexes (IDs) to the columns on the input buffer
         // Work backwards by matching the batchTable column name with the ExternalOutput column name
         //   then get the ID from the ExternalOutput. Match the ExternalOutput ID with the Input column 
         //   External Column ID and get the LineageID, which will be the ID of the buffer column.
         //   Set the col index collection value for the batchTable column to the buffer ID.

         // Create a mapping from external column name to ID
         // Create a mapping from output column's referenced external column
         //    ID to the columns lineage ID.
         // For each column in the batchTable, find the external column and from
         //    that find the lineage ID.
         //

         // Get the input and the external column collection
         IDTSInput100 input = ComponentMetaData.InputCollection[0];
         IDTSExternalMetadataColumnCollection100 externalcols = input.ExternalMetadataColumnCollection;

         // Make a hash table for the names of the external columns and contains the ID
         int cExternalCol = externalcols.Count;
         System.Collections.Hashtable hashExternalNameToID = new System.Collections.Hashtable(cExternalCol);

         // Build a mapping between the external column name and its ID
         for (int iExternalCol = 0; iExternalCol < cExternalCol; iExternalCol++)
         {
            IDTSExternalMetadataColumn100 externalcol = externalcols[iExternalCol];
            hashExternalNameToID.Add(externalcol.Name, externalcol.ID);
         }

         // Get the input columns collection and make a hash table to map the external
         // ID that it references to the lineage ID of the column
         IDTSInputColumnCollection100 inputcols = input.InputColumnCollection;
         int cInputCol = inputcols.Count;
         System.Collections.Hashtable hashReferenceIDToLineageID = new System.Collections.Hashtable(cInputCol);

         // Build a mapping between external columns and input buffer columns IDs
         for (int iInputCol = 0; iInputCol < cInputCol; iInputCol++)
         {
            IDTSInputColumn100 inputcol = inputcols[iInputCol];
            if (inputcol.ExternalMetadataColumnID > 0)
               hashReferenceIDToLineageID.Add(inputcol.ExternalMetadataColumnID, inputcol.LineageID);
         }

         // Loop over all columns in the batchTable
         for (int iRSCol = 0; iRSCol < cBTCol; iRSCol++)
         {
            // Get the name for the column at the current index
            string strRSColName = m_batchTable.Columns[iRSCol].ColumnName;
               
            // Find the ID of the external column from the name
            // ExternalMetaData is driven from the destination table.
            // This should not fail unless ExternalMetadata is corrupt.
            int idExternal = (int)hashExternalNameToID[strRSColName];

            // Find the ExternalMetadata column that points to this ID
            try
            {
               int linid = (int)hashReferenceIDToLineageID[idExternal];
               // a hit, so get the index in the buffer of this column
               int idx = (int)BufferManager.FindColumnByLineageID(input.Buffer, linid);
               m_idxColumn[iRSCol] = idx;
            }
            catch (Exception)
            {
               // didn't find the column name in the input collection, so this column gets a null
               m_idxColumn[iRSCol] = -1;
            }
         }
      }

      // Take the data from the buffer and put it into the adapter.
      // After each buffer, write the data in the dataset to the destination
      public override void ProcessInput(int inputID, PipelineBuffer buffer)
      {
         uint uiRowsWritten = 0;
         uint uiBlobBytesWritten = 0;

         int cCols = ComponentMetaData.InputCollection[0].ExternalMetadataColumnCollection.Count;
         try
         {
            while (buffer.NextRow())
            {
               // collect an array of values from the pipeline
               for (int iCol = 0; iCol < cCols; iCol++)
               {
                  // Which buffer column is this ?
                  int idxCol = m_idxColumn[iCol];
                                    
                  // if no input is mapped or the value in the buffer is null, then set null to the array
                  if (idxCol == -1)
                  {
                     m_odbcAdapter.InsertCommand.Parameters[iCol].Value = null;
                  }
                  else if (buffer.IsNull(idxCol))
                  {
                     m_odbcAdapter.InsertCommand.Parameters[iCol].Value = DBNull.Value;
                  }
                  else
                  {
                     // Set the parameter value for the insert.                     
                     m_odbcAdapter.InsertCommand.Parameters[iCol].Value = buffer[idxCol];
                                        
                     // Specially treat BLOB columns
                     BlobColumn blob = buffer[idxCol] as BlobColumn;
                     if (blob != null)
                     {
                        DataType dataType = blob.ColumnInfo.DataType;
                        if (dataType == DataType.DT_TEXT
                           || dataType == DataType.DT_NTEXT)
                           m_odbcAdapter.InsertCommand.Parameters[iCol].Value = buffer.GetString(idxCol);
                        else if (dataType == DataType.DT_IMAGE)
                           m_odbcAdapter.InsertCommand.Parameters[iCol].Value = blob.GetBlobData(0, (int)blob.Length);

                        try
                        {
                           // Update the number of blob bytes written
                           uiBlobBytesWritten = checked(uiBlobBytesWritten + blob.Length);
                        }
                        catch (System.OverflowException)
                        {
                           // Overflow so tell the perfcounter how much we
                           // have written so far
                           ComponentMetaData.IncrementPipelinePerfCounter(DTS_PIPELINE_CTR_BLOBBYTESWRITTEN, uiBlobBytesWritten);

                           // Keep track of the new amount that would have
                           // caused the overflow
                           uiBlobBytesWritten = blob.Length;
                        }
                     }
                  }
               }

               int hr = 0;

               try
               {
                  int nRows = m_odbcAdapter.InsertCommand.ExecuteNonQuery();
                  m_odbcAdapter.Update(m_batchTable);

                  // Mark that we added a row
                  uiRowsWritten++;
               }
               catch (Exception e)
               {
                  if (m_rowdisp == DTSRowDisposition.RD_FailComponent)
                     throw e;

                  if (m_rowdisp == DTSRowDisposition.RD_IgnoreFailure)
                     continue;

                  if (m_rowdisp == DTSRowDisposition.RD_RedirectRow)
                  {
                     buffer.DirectErrorRow(m_errorOutID, hr, 0);
                  }
               }
            }
         }
         finally
         {
            // Update the number of rows we inserted
            ComponentMetaData.IncrementPipelinePerfCounter(DTS_PIPELINE_CTR_ROWSWRITTEN, uiRowsWritten);

            // Update the number of blob bytes we written
            ComponentMetaData.IncrementPipelinePerfCounter(DTS_PIPELINE_CTR_BLOBBYTESWRITTEN, uiBlobBytesWritten);
         }
      }
      #endregion
   }
}

